##
## Material adaptado de cdigos orinais de S. Wood
##

par.ori <- par(no.readonly=TRUE)

#####################################################################
## 1. Preciso numrica - um exemplo com regresso polinomial
#####################################################################

## simulando dados
set.seed(1)
n <- 100
x1 <- sort(runif(n))
y <- 0.2*(x1-0.5) + (x1-0.5)^2 + rnorm(n, sd=0.1)
x1
y

plot(y ~ x1)

## I) modelo de regresso polinomial em x1
## i) usando a funo lm()
reg1 <- lm(y ~ x1 + I(x1^2))
reg1

## preditos no grfico
x.pred <- data.frame(x1 = seq(0,1,length=200))
lines(x.pred$x1, predict(reg1, newdata=x.pred))

## (ii) fazendo por operaes matriciais
X <- cbind(1, x1, I(x1^2))
X

## OBS: formas alternativas/equivalentes (e mais interessantes!) de montar a matriz X
##X <- model.matrix(reg) 
##X <- model.matrix(~ x1+I(x1^2)) 

## estimativa de minimos quadrados
(beta.hat <- solve(t(X)%*%X) %*%  t(X)%*%y)  ## no devemos inverter a matrix .... 
(beta.hat <- solve(t(X)%*%X, t(X)%*%y))      ##  melhor resolver o sistema!

## os resultados so or mesmos
coef(reg1)
drop(beta.hat)


## II) regresso em x2 obtido simplesmente deslocando os valores de x1
x2 <- x1 + 100
reg2 <- lm(y ~ x2 + I(x2^2))
reg2

## mas tentando fazer matricialmente agora (ingenuamente...)
X <- model.matrix(~ x2 + I(x2^2)) 
beta.hat <- solve(t(X)%*%X, t(X)%*%y) 
## ou de outra forma...
beta.hat <- solve(crossprod(X), crossprod(X,y))

## Portanto no obtemos a soluo por operaes matriciais

## OBS:  possvel capturar e tratar o erro com mensagens (pode ser til em programao)
## evitando o erro (util para manipulao de opes dentro de funes)
beta.hat <- try(solve(crossprod(X), crossprod(X,y)), silent=T)
beta.hat
if(class(beta.hat) == "try-error")
  beta.hat <- "SOLUES ALTERNATIVAS AQUI (VER MAIS ADIANTE)"

## Portanto neste caso lm() forneceu estimativas mas as operaes matriciais no!
## Uma possvel soluo:
## reescalonar as variveis, estimar coeficientes modificados e depois obter os origianais

## III) regresso em x3 obtido deslocando ainda mais os valores de x1
## mas mesmo lm() pode falhar ou produzir resultados estranhos! (mesmo que no aparente isto!!!)
x3 <- x1 + 1000
reg3 <- lm(y ~ x3 + I(x3^2))
reg3 

## note que a funo roda com um "erro silencioso"

## e compare os modelos ajustados nos grficos
par(mfrow=c(1,3))
plot(x1, y)
#xvals <- seq(min(x1), max(x1), len=100)
#lines(xvals, predict(reg1, newdata=data.frame(x1 = xvals)))
lines(x1, fitted(reg1))
##
plot(x2, y)
#xvals <- seq(min(x2), max(x2), len=100)
#lines(xvals, predict(reg2, newdata=data.frame(x2 = xvals)))
lines(x2, fitted(reg2))
##
plot(x3, y)
#xvals <- seq(min(x3), max(x3), len=100)
#lines(xvals, predict(reg3, newdata=data.frame(x3 = xvals)))
lines(x3, fitted(reg3))

## O grfico do modelo reg3 mostra predies inadequadas obtidas com as soluo absurda
par(par.ori)

## Investigar tb o comportamento usando poly() para definir a regresso polinomial!!!

## por default  feita uma ortonormalizao
reg4 <- lm(y ~ poly(x3, degree=2))
reg4
## ... e portanto os coeficientes so "ajustados"
## mas isto no afeta os valores preditos!
plot(x3, y)
#lines(xvals, predict(reg4, newdata=data.frame(x3 = xvals)))
lines(x3, fitted(reg4))

## o equivalente ao anterior seria:
reg4r <- lm(y ~ poly(x3, degree=2, raw=TRUE))
reg4r
## que tambm causas problemas
lines(x3, fitted(reg4r))


#####################################################################
## 2. Eficincia em clculo matricial
#####################################################################
n <- 1000
A <- matrix(runif(n^2), n, n)
B <- matrix(runif(n*n), n, n)
y <- runif(n)
system.time(A%*%B%*%y) 
system.time(A%*%(B%*%y))

## verificar nmero de "flops" envolvidos nas operaes!!

## Trao de matrix
n <- 5000
m <- 100
A <- matrix(runif(n*m), n, m)
B <- matrix(runif(n*m), m, n)
system.time(sum(diag(A%*%B))) # 2n^2m
system.time(sum(diag(B%*%A))) # 2nm^2 ( tr(AB) = tr(BA) )
system.time(sum(A*t(B)))      # 2nm   (tr(AB) = sum_ij Aij Bji


#####################################################################
## 3. Decomposio de Cholesky
#####################################################################
## S = R'R ;   R  triangular superior

## 3.1 Simulao de NMV: y = mu + R'z
S <- matrix(c(5,-3, 2, -3,5,-3, 2,-3,5), 3, 3)
S
mu <- c(1, 0, -4)
R <- chol(S) 
R
t(R)
z <- rnorm(3) 
y <- mu + t(R)%*%z
y
y <- mu + crossprod(R,z)
y

## verificar/ilustrar comportamento de mdias... varincias e correlaes!
z <- matrix(rnorm(30), nc=10) 
y <- mu + t(R)%*%z
## que  equivalente a:
y <- mu + crossprod(R, z)
round(y, dig=3)

## comparando parmetros verdadeiros e estimativas das simulaes
## simulando agora 10,000 vetores
y <- mu + crossprod(R, matrix(rnorm(30000), nc=10000))
dim(y)

# mdias
mu          ## terica
rowMeans(y) ## mdia dos valores simulados
# varincias
diag(S)           ## terica
apply(y, 1, var)  ## varincia dos valores simulados

# correlaes
Cor <- S/outer(diag(S), diag(S), function(x,y) sqrt(x)*sqrt(y))
Cor[lower.tri(Cor)]
cor.d <- cor(t(y))      ## terica
cor.d[lower.tri(cor.d)] ## correlao dos valores simulados

## 3.2 Uso no clculo de verossimilhana (modelo normal mv)
y <- 1:3
## expresso da log-verossimilhana do modelo normal multivariado
## l(\theta) = -(n/2) log(2*\pi) - (1/2) log(|\Sigma|) - (1/2) (y - F\beta)'\Sigma^{-1} (y - F\beta)

## clculo "ingnuo"
logLik <- -0.5*(length(y) * log(2*pi) + determinant(S, log=TRUE)$mod + crossprod((y - mu), solve(S, (y - mu))))
logLik

## e a verossimilhana pode ser calculada mais adequadamente por:
z <- forwardsolve(t(R), y-mu)
z
logLik <- -length(y)*log(2*pi)/2 - sum(log(diag(R))) - sum(z*z)/2
logLik
## notar que det(S) = (prod(R_ii))^2

########################################
## 4. Outras decomposies
########################################
set.seed(123)
x1 <- sort(runif(100))
x <- x1 + 100 
X <- model.matrix(~ x + I(x^2)) 

## 4.1 Autovalores e autovetores
lambda <- eigen(t(X)%*%X)$values
lambda
lambda <- eigen(crossprod(X))$values
lambda
lambda[1]/lambda[3]

## 4.2 Decomposio de valor singular
d <- svd(X)$d 
d
d[1]/d[3]


## 4.3 Decomposio QR: X = QR,  R tring sup e Q'Q = Im (mas QQ' != I)
X.qr <- qr(X) 

## Exerccio: revisitar soluo do sistema de equaes normais
## e o clculo da funo de verossimilhana da normal multivariada utilizando:
##   (i) a decomposio spectral,
##  (ii) decomposio de valores singulares e
##  (iii) decomposio QR


########################################################################
## 6. Integrao Numrica
########################################################################

## quadraturas -- funo escada
# ex: \int_0^1 exp(x) dx

## soluo analtica:
I <- exp(1) - 1
I

## soluo por integrate() (uma dimenso e limitada)
I.int <- integrate(function(x) exp(x), lower=0, upper=1)
I.int
c(I.int$val, I, (I-I.int$val)/I)

## approximao por ponto mdio
N <- 10
I.pm <- sum(exp((1:N-.5)/N))/N
I.pm
c(I.pm, I, (I-I.pm)/I)

## pode-se aumentar a acurcia...
N <- 100
I.pm <- sum(exp((1:N-.5)/N))/N
c(I.pm, I, (I-I.pm)/I)

N <- 1000
I.pm <- sum(exp((1:N -.5)/N))/N
c(I.pm, I, (I-I.pm)/I)

## quadratura gaussiana
require(statmod)
N <- 10
## pontos x_i de integrao e pesos w_i
qg <- gauss.quad(N) 
qg
## a funo  reescrita para aplicar a quadratura e no intervalo [-1,1]
## usando a relao \int^a_b f(x) dx = \int_{-1}^1 f(/frac{a+b+(b-a)t}{2}) dt
I.qg <- sum(qg$weights * exp((qg$nodes+1)/2))/2
c(I.qg,I,(I-I.qg)/I)

## reduzindo o nmero de pontos...
N <- 5
gq <- gauss.quad(N) 
I.qg <- sum(gq$weights*exp((gq$nodes+1)/2))/2
c(I.qg, I, (I-I.qg)/I)

N <- 3
qg <- gauss.quad(N) 
I.qg <- sum(qg$weights*exp((qg$nodes+1)/2))/2
c(I.qg, I, (I - I.qg)/I)

## examinar os argumentos da funo para outras opes e mtodos! 

## outras dadas opes so ainda dadas em gauss.quad.prob()
args(gauss.quad.prob)

## uniforme
qgu <- gauss.quad.prob(10, dist="uniform", l=0, u=1)
qgu
I.qgu <- sum(qgu$w * exp(qgu$n))
I.qgu
c(I.qgu, I, (I - I.qgu)/I)

## normal (no  muito natural para este exemplo, pode-se tentar
## diferentes opes para mu e sigma mas aproximao no funciona muito bem...)
## Tem-se aqui o problema do domnio de integrao estar restrito de em [0,1] 
## Esta opo  mais adequada para variveis como domnio de integrao entre (-Inf, Inf) 
## e requer que mu e sigma sejam informados
qgn <- gauss.quad.prob(10, dist="normal", mu=0.5, sigma=0.2)
qgn

## idem anterior para Gamma com parmetro de forma alpha e escala beta
## s que neste caso com intervalo (0, +Inf)
qgg <- gauss.quad.prob(10, dist="gamma", alpha=1, beta=0.5)
qgg
## e note que para alpha (shape) = 1 tem-se a exponencial
ff <- function(x) dgamma(x, sh=1, sc=1/5)
curve(ff, from=0, to=3)
#
dgamma(0:5, sh=1, rate=1/10)
dexp(0:5, rate=1/10)

## Outra alternativa para o intervalo (0, 1)  dado pela Beta
##
bb <- function(x, a, b) x^(a-1)*(1-x)^(b-1)/beta(a,b)
bb((0:10)/10, 2, 3)
dbeta((0:10)/10, 2, 3)
ff <- function(x) dbeta(x, 2, 2)
curve(ff, 0, 1)
ff <- function(x) dbeta(x, 0.5, 0.5)
curve(ff, 0, 1, add=T, lty=2)
ff <- function(x) dbeta(x, 1.5, 0.5)
curve(ff, 0, 1, add=T, lwd=2)
ff <- function(x) dbeta(x, 1, 1)
curve(ff, 0, 1, add=T, lty=2, lwd=2)

## mas note que a qualidade a aproximao depende da escolha dos parmetros
qgb <- gauss.quad.prob(10, dist="beta", a=2, b=2)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=0.5, b=0.5)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=1.5, b=0.5)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

qgb <- gauss.quad.prob(10, dist="beta", a=1, b=1)
qgb
I.qgb <- sum(qgb$w * exp(qgb$n))
I.qgb
c(I.qgb, I, (I - I.qgb)/I)

##
## Integrao Multidimensional...
##

## baseado na aplicao recursiva de regras de 1-D
## \int \int f(x,y) dx dy \approx \sum_j \sum_i w_j w_i f(x_j, y_i)

## funo para encontrar conjunto de ns e pesos d-dimensional,
malha <- function(nos, dim, pesos = 1/length(x)+x*0) {
  ## nos: ns (em 1-D) onde a funo deve ser avaliada
  ## dim: dimenso do espao de integrao
  ## pesos: pesos correspondentes aos ns 
  n <- length(nos)
  X <- W <- matrix(0, n^dim, dim)
  for (i in 1:dim) {
    X[,i] <- nos
    W[,i] <- pesos
    nos <- rep(nos, rep(n, length(nos)))
    pesos <- rep(pesos, rep(n, length(pesos)))
  }
  ## pesos so dados pelo produto das colunas de W
  pesos <- exp(rowSums(log(W))) 
  ## e cada linha de X fornece as coordenadas dos ns
  list(nodes=X, weights=pesos) 
}

## Ex: integrar f(x) = exp(\sum x_i), em [-1; 1]^d
fd <- function(x) {exp(rowSums(x))}

N <- 10; d <- 3
## nmero de avaliaes da funo:
N^d

## soluo analtica
I <- (exp(1)-exp(-1))^d

## Aproximao por ponto mdio
malha.pm <- malha((1:N-.5)/N*2-1, d, rep(2/N,N))
I.pm <- sum(malha.pm$weights * fd(malha.pm$nodes))
c(I.pm, I,(I-I.pm)/I)

## agora em 6-D 
N <- 10; d <- 6
## nmero de avaliaes da funo
N^d

I <- (exp(1)-exp(-1))^d
malha.pm <- malha((1:N-.5)/N*2-1, d, rep(2/N,N))
I.pm <- sum(malha.pm$wei * fd(malha.pm$nodes))
c(I.pm, I,(I-I.pm)/I)


## agora por quadratura gaussiana, aumenta (mas nem tanto!) o nmero de avaliaes da funo
N <- 5; d <- 2
N^d

I <- (exp(1)-exp(-1))^d
gq <- gauss.quad(N)
malha.qg <- malha(gq$nodes, d, gq$weights)
I.qg <- sum(malha.qg$wei * fd(malha.qg$nodes))
c(I.qg, I, (I-I.qg)/I)

## comparar o efeito do aumento da dimenso
N <- 10; d <- 6
N^d

I <- (exp(1)-exp(-1))^d
qg <- gauss.quad(N)
malha.gq <- malha(gq$nodes, d, gq$weights)
I.gq <- sum(malha.qg$wei * fd(malha.qg$nodes))
c(I.qg, I, (I-I.qg)/I)


## experimentar com outros nmeros de ns etc etc
                                                     
## Aproximando o integrando: abordagem por Aproximao de Laplace

## Integrao de Monte Carlo

## Monte Carlo estratificado

## Amostragem por importncia (Importance sampling)

## Amostragem por importncia via Laplace (Laplace importance sampling)


